@@ -16,12 +16,21 @@ import pickle
 from . import infin_ast
 from datetime import datetime, timedelta
 import time
-from mlflow import start_run, end_run, log_metric, log_param, log_artifacts, set_experiment
+from mlflow import start_run, end_run, log_metric, log_param, log_artifacts, set_experiment, log_artifact
 from os.path import sep
 import mlflow.projects
 from contextlib import contextmanager,redirect_stderr,redirect_stdout
 import pprint
 import types
+import shutil
+from . import servicedefs
+import builtins
+from infinstor_mlflow_plugin.tokenfile import read_token_file
+from os.path import expanduser
+from requests.exceptions import HTTPError
+import requests
+import inspect
+import glob
 
 # import astpretty
 
@@ -30,6 +39,7 @@ TRANSFORM_RAW_DS = "infin_transform_raw_to_ds"
 TRANSFORM_CSV_PD = "infin_transform_csv_to_pd"
 TRANSFORM_CSV_DS = "infin_transform_csv_to_ds"
 TRANSFORM_ONE_OBJ = "infin_transform_one_object"
+TRANSFORM_DIR_BY_DIR = "infin_transform_dir_by_dir"
 
 verbose = False
 
@@ -45,6 +55,59 @@ class FuncLister(ast.NodeVisitor):
         # print('>> Funclister: ' + node.name)
         self.generic_visit(node)
 
+def get_label_info(label):
+    tokfile = expanduser('~') + sep + '.infinstor' + sep + '/token'
+    token, refresh_token, token_time, client_id, service = read_token_file(tokfile)
+
+    payload = ("label=" + label)
+    headers = {
+        'Content-Type': 'application/x-www-form-urlencoded',
+        'Authorization': token
+        }
+    url = 'https://api.' + builtins.service + '.com/listlabels'
+    try:
+        response = requests.post(url, data=payload, headers=headers)
+        response.raise_for_status()
+    except HTTPError as http_err:
+        print(f'HTTP error occurred: {http_err}')
+        raise
+    except Exception as err:
+        print(f'Other error occurred: {err}')
+        raise
+    labels = response.json()['labels']
+    for one_label in labels:
+        labelname = one_label['label']
+        if (labelname == label):
+            return one_label['timespec'], one_label['bucketname'], one_label['prefix']
+        else:
+            return None, None, None
+
+
+def download_one_dir(label):
+    timespec, bucketname, prefix = get_label_info(label)
+    endpoint = "https://lbl" + label + ".s3proxy." + builtins.service + ".com:443/";
+
+    session = boto3.session.Session(profile_name='infinstor')
+    client = session.client('s3', endpoint_url=endpoint)
+    dict_of_arrays_of_files = dict()
+    list_dir_by_dir(client, bucketname, prefix, True, dict_of_arrays_of_files)
+    tmpdir_root = tempfile.mkdtemp()
+    try:
+        start_run()
+        for parentdir in dict_of_arrays_of_files:
+            array_of_files = dict_of_arrays_of_files[parentdir]
+            if (verbose == True):
+                info('Number Of Objects in parentdir ' + parentdir + ': ' +str(len(array_of_files)))
+            local_tmpdir = tmpdir_root + sep + parentdir
+            if (verbose == True):
+                info('Local temp dir is ' + local_tmpdir)
+            os.makedirs(local_tmpdir, mode=0o755, exist_ok=True)
+            objects = download_objects_inner(endpoint, bucketname, parentdir, array_of_files,\
+                    False, None, local_tmpdir)
+    finally:
+        end_run()
+    return tmpdir_root
+
 # fills out array_of_files with all the files in this prefix
 def list_one_dir(client, bucket, prefix_in, recurse, array_of_files):
     paginator = client.get_paginator('list_objects_v2')
@@ -124,6 +187,15 @@ def info(msg):
     now = datetime.now()
     print(__name__ + '[' + str(os.getpid()) + '][' + now.strftime('%Y-%m-%d %H:%M:%S') + ']' + msg)
 
+def log_artifacts_recursively(temp_output_dir):
+    for filename in glob.iglob(temp_output_dir + '**/**', recursive=True):
+        if (os.path.isfile(filename)):
+            bn = os.path.basename(filename)
+            bn_minus_tmp = filename[len(temp_output_dir):].lstrip()
+            key = 'infinstor/' + os.path.dirname(bn_minus_tmp).lstrip('/')
+            info("Logging artifact " + str(filename) + " to " + key)
+            log_artifact(filename, key)
+
 class download_task:
     def __init__(self, command, bucketname, parentdir, filename):
         self.command = command
@@ -131,8 +203,19 @@ class download_task:
         self.parentdir = parentdir
         self.filename = filename
 
-def s3downloader(download_task_q, write_pipe, endpoint, glb):
-    # info('s3downloader: started. endpoint=' + endpoint)
+def s3downloader(**kwargs):
+    download_task_q = kwargs.pop('command_q')
+    write_pipe = kwargs.pop('write_pipe')
+    endpoint = kwargs.pop('endpoint')
+    glb = kwargs.pop('globals')
+    tmpfile_dir = kwargs.pop('tmpfile_dir')
+
+    # 1. tmpfile_dir=None, glb=None: Read object into memory and return bytes to pipe
+    # 2. tmpfile_dir is present and glb=None: Read object into tmpfile_dir and return status
+    # 3. tmpfile_dir is present and glb is present: Read object into tmpfile_dir and call
+    #    the function infin_transform_one_object using glb as globals. Send status back
+    # Note: If glb is present, tmpfile_dir must be present. We always call
+    # infin_transform_one_object by passing in a temporary file
 
     if (glb != None):
         namespaced_infin_transform_one_object =\
@@ -152,11 +235,14 @@ def s3downloader(download_task_q, write_pipe, endpoint, glb):
                 if (ms > 10):
                     info('s3downloader: q.get took ' + str(ms) + ' ms')
         except Empty as e:
-            info('s3downloader: No more entries in download_task_q. Exiting..')
+            if (verbose == True):
+                info('s3downloader: No more entries in download_task_q. Exiting..')
             write_pipe.close()
             break
         op = download_task.command
         if (op == 'download'):
+            if (verbose == True):
+                info('s3downloader: Received download command')
             bucketname = download_task.bucketname
             parentdir = download_task.parentdir
             filename = download_task.filename
@@ -168,44 +254,55 @@ def s3downloader(download_task_q, write_pipe, endpoint, glb):
                 info('s3downloader: starting download of ' + full_object_key\
                         + ' from ' + bucketname)
                 start = datetime.now()
-            obj = client.get_object(Bucket=bucketname, Key=full_object_key)
-            content_length = obj['ContentLength']
-            if (verbose == True):
-                end = datetime.now()
-                info('s3downloader: finished download of ' + full_object_key\
-                        + ' from ' + bucketname)
-                td = end - start
-                ms = (td.days * 86400000) + (td.seconds * 1000) + (td.microseconds / 1000)
-                if (ms > 500):
-                    info('s3downloader: download of size ' + str(content_length) + ' took '\
-                        + str(ms) + ' ms')
 
-            strbody = obj['Body']
-            dtm = obj['LastModified']
-            bts = strbody.read()
-            key = dtm.strftime('%Y-%m-%d %H:%M:%S') + ' ' + bucketname + '/' + full_object_key
+            if (tmpfile_dir == None):
+                obj = client.get_object(Bucket=bucketname, Key=full_object_key)
+                content_length = obj['ContentLength']
+                strbody = obj['Body']
+                bts = strbody.read()
+                dtm = obj['LastModified']
+                key = dtm.strftime('%Y-%m-%d %H:%M:%S') + ' ' + bucketname + '/' + full_object_key
+            else:
+                tmpf_name = tmpfile_dir + sep + filename
+                client.download_file(bucketname, full_object_key, tmpf_name)
+                content_length = os.stat(tmpf_name).st_size
+                key = bucketname + '/' + full_object_key
+
             if (glb == None):
-                if (not write_pipe.writable):
+                while (not write_pipe.writable):
                     info('s3downloader: WARNING write pipe not writable')
-                    time.sleep(1)
+                    time.sleep(2)
                 try:
                     write_pipe.send(key)
-                    write_pipe.send(bts)
+                    write_pipe.send(filename)
+                    if (tmpfile_dir == None):
+                        write_pipe.send(bts)
+                    else:
+                        write_pipe.send('Success')
                 except Exception as e:
                     status_str = str(e)
                     info("Error sending bytes back: " + status_str)
             else:
                 try:
-                    namespaced_infin_transform_one_object(bucketname, parentdir, filename, bts)
-                except Exception as e:
-                    status_str = str(e)
-                    info("Error executing infin_transform_one_object: " + status_str)
+                    temp_output_dir = tempfile.mkdtemp()
+                    #namespaced_infin_transform_one_object(bucketname, parentdir,\
+                    #        filename, tmpf_name, **kwargs)
+                    namespaced_infin_transform_one_object(tmpf_name, temp_output_dir, **kwargs)
+                except Exception as e1:
+                    status_str = str(e1)
+                    info("Error executing infin_transform_one_object_tmpfile: " + status_str)
                 else:
+                    log_artifacts_recursively(temp_output_dir)
                     status_str = 'Success'
+                finally:
+                    shutil.rmtree(temp_output_dir)
+                os.remove(tmpf_name)
                 write_pipe.send(key)
+                write_pipe.send(filename)
                 write_pipe.send(status_str)
         elif (op == 'quit'):
-            info('s3downloader: Received quit')
+            if (verbose == True):
+                info('s3downloader: Received quit command')
             write_pipe.close()
             break
         else:
@@ -246,14 +343,15 @@ def load_one_csv_from_bytearray(bts):
 def download_objects(endpoint, bucketname, parentdir, array_of_files, is_csv, glb):
     try:
         start_run()
-        return download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_csv, None)
+        return download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_csv, None, None)
     finally:
         end_run()
 
 # returns a pandas DataFrame with index 'YY-MM-dd HH:MM:SS bucketname/filename'
 # and one column named RawBytes that contains the raw bytes from the object
 # This inner method is called after the outer method wraps in mlflow start_run
-def download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_csv, glb):
+def download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_csv,\
+        glb, tmpfile_dir, **kwargs):
     log_param("object_count", len(array_of_files))
 
     command_q = Queue(len(array_of_files) + num_threads())
@@ -267,12 +365,19 @@ def download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_c
     for i in range(num_threads()):
         r1, w1 = Pipe(False)
         pipe_from_child.append(r1)
-        p = Process(target=s3downloader, args=(command_q, w1, endpoint, glb,))
+        newkwargs = dict(kwargs)
+        newkwargs['command_q'] = command_q
+        newkwargs['write_pipe'] = w1
+        newkwargs['endpoint'] = endpoint
+        newkwargs['globals'] = glb
+        newkwargs['tmpfile_dir'] = tmpfile_dir
+        p = Process(target=s3downloader, args=(), kwargs=newkwargs)
         p.start()
         processes.append(p)
         w1.close()
 
     filebytes = []
+    filenames = []
     filekeys = []
     step = 0
     with tqdm(total=len(array_of_files)) as pbar:
@@ -283,8 +388,10 @@ def download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_c
                 try:
                     while (read_pipe.poll()):
                         key = read_pipe.recv()
+                        fn = read_pipe.recv()
                         bts = read_pipe.recv()
                         filekeys.append(key)
+                        filenames.append(fn)
                         filebytes.append(bts)
                         pbar.update(1)
                         files_read += 1
@@ -299,29 +406,27 @@ def download_objects_inner(endpoint, bucketname, parentdir, array_of_files, is_c
     if (is_csv == True):
         rv = pd.concat(map(load_one_csv_from_bytearray, filebytes))
     else:
-        data = {'RawBytes': filebytes}
+        data = {'FileName': filenames, 'RawBytes': filebytes}
         rv = DataFrame(data, index=filekeys)
     log_metric("downloaded", files_read, step=step)
     return rv
 
 def actually_run_transformation(client, endpoint, cache_path, is_pandas_df, bucketname,\
-        prefix_in, is_csv, xformname):
-    xform_obj = client.get_object(Bucket='infinstor-pseudo-bucket', Key='transforms/' + xformname)
-    strbody = xform_obj['Body']
-    bts = strbody.read()
-    s = str(bts, 'utf-8')
-    # print(s)
-
+        prefix_in, is_csv, xformname, transformation_string, one_object_only, **kwargs):
     array_of_files = []
     list_one_dir(client, bucketname, prefix_in, True, array_of_files)
-    print('Total Number Of Objects: ' + str(len(array_of_files)))
+    if (one_object_only):
+        first = array_of_files[0]
+        array_of_files = []
+        array_of_files.append(first)
+    print('actually_run_transformation: total number Of objects: ' + str(len(array_of_files)))
     try:
         start_run()
-        objects = download_objects_inner(endpoint, bucketname, '', array_of_files, is_csv, None)
+        objects = download_objects_inner(endpoint, bucketname, '', array_of_files, is_csv, None, None, **kwargs)
     finally:
         end_run()
 
-    transformAst = infin_ast.extract_transform(TRANSFORM_RAW_PD, src_str=s)
+    transformAst = infin_ast.extract_transform(TRANSFORM_RAW_PD, src_str=transformation_string)
     # XXX we should use the following statement to figure out what
     # kind of an object the infin_transform function returns
     # infin_ast.add_type_statements(transformAst)
@@ -341,34 +446,60 @@ def actually_run_transformation(client, endpoint, cache_path, is_pandas_df, buck
         else:
             xn = TRANSFORM_RAW_DS
 
-    transformSrc = transformSrc + "\n" + xn + "(objects)\n"
-
     tree = ast.parse(transformSrc)
+
+    compiledcode2 = compile(tree, "<string>", "exec")
+
     # Add all functions in xformcode to the globals dictionary
     glb = {}
     fl = FuncLister(glb)
     fl.visit(tree)
-    glb['objects'] = objects;
-    #astpretty.pprint(tree)
-    compliedcode = compile(tree, "<string>", "exec")
-    exec(compliedcode, glb)
-    # save result of transformation in cache
+    try:
+        exec(compiledcode2, glb)
+    except Exception as e:
+        status_str = str(e)
+    else:
+        status_str = 'Success'
+    # print('Globals=')
+    # pprint.pprint(glb)
+    info("Execution of global statics complete. status_str=" + status_str)
+
+    namespaced_infin_transform_fnx = namespaced_function(glb[xn], glb, None, True)
+    try:
+        namespaced_infin_transform_fnx(objects, **kwargs)
+    except Exception as e:
+        status_str = str(e)
+        info("Error executing " + xn + ", status=" + status_str)
+    else:
+        status_str = 'Success'
+
     if (is_pandas_df == True):
-        fd, tmpf_name = tempfile.mkstemp(suffix='.pkl')
-        os.close(fd)
-        objects.to_pickle(tmpf_name)
-        client.put_object(Body=open(tmpf_name, 'rb'), Bucket='infinstor-pseudo-bucket',\
+        if (cache_path != None):
+            fd, tmpf_name = tempfile.mkstemp(suffix='.pkl')
+            os.close(fd)
+            objects.to_pickle(tmpf_name)
+            client.put_object(Body=open(tmpf_name, 'rb'), Bucket='infinstor-pseudo-bucket',\
                 Key=cache_path)
-        os.remove(tmpf_name)
+            os.remove(tmpf_name)
+        else:
+            print('cache_path not specified. not caching')
     else:
         print('saving tf.data.Dataset unimplemented')
     return objects
 
-def read_raw_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, is_csv, xformname):
+def read_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, is_csv,\
+        xformname, transform_string, one_object_only, **kwargs):
     endpoint = "https://" + time_spec + ".s3proxy." + service_name + ".com:443/";
     session = boto3.session.Session(profile_name='infinstor')
     client = session.client('s3', endpoint_url=endpoint)
 
+    cache_path = None
+    if (xformname == None):
+        print('xformname not specified. Not checking cache')
+        return actually_run_transformation(client, endpoint, cache_path, True,\
+                bucketname, prefix_in, is_csv, xformname, transform_string,\
+                one_object_only, **kwargs)
+
     # check if we can download a cached version
     cache_path = 'cache/' + xformname + '/' + time_spec + '/'\
             + bucketname + '/' + prefix_in + 'pd.DataFrame'
@@ -381,15 +512,22 @@ def read_raw_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, is_
     except client.exceptions.NoSuchKey:
         print('cache miss for ' + cache_path)
         cached_obj = actually_run_transformation(client, endpoint, cache_path, True,\
-                bucketname, prefix_in, is_csv, xformname)
+                bucketname, prefix_in, is_csv, xformname, transform_string, False, **kwargs)
     finally:
         return cached_obj
 
-def read_raw_and_xform_to_ds(time_spec, service_name, bucketname, prefix_in, is_csv, xformname):
+def read_and_xform_to_ds(time_spec, service_name, bucketname, prefix_in, is_csv,\
+        xformname, transform_string, one_object_only, **kwargs):
     endpoint = "https://" + time_spec + ".s3proxy." + service_name + ".com:443/";
     session = boto3.session.Session(profile_name='infinstor')
     client = session.client('s3', endpoint_url=endpoint)
 
+    if (xformname == None):
+        print('xformname not specified. Not checking cache')
+        return actually_run_transformation(client, endpoint, cache_path,\
+                False, bucketname, prefix_in, is_csv,\
+                xformname, transform_string, one_object_only, **kwargs)
+
     # check if we can download a cached version
     cache_path = 'cache/' + xformname + '/' + time_spec + '/'\
             + bucketname + '/' + prefix_in + 'tf.data.Dataset'
@@ -401,40 +539,54 @@ def read_raw_and_xform_to_ds(time_spec, service_name, bucketname, prefix_in, is_
     except client.exceptions.NoSuchKey:
         print('cache miss for ' + cache_path)
         cached_obj = actually_run_transformation(client, endpoint, cache_path, False,\
-                bucketname, prefix_in, is_csv, xformname)
+                bucketname, prefix_in, is_csv, xformname, transform_string, False, **kwargs)
     finally:
         return cached_obj
 
-def run_xform_periodically(seconds, service_name, bucketname, prefix_in, xformname):
-    while (True):
-        start_time = datetime.utcnow()
-        time.sleep(seconds)
-        end_time = datetime.utcnow()
-        time_spec = 'tm' + start_time.strftime("%Y%m%d%H%M%S")\
-                + '-tm' + end_time.strftime("%Y%m%d%H%M%S")
-        read_raw_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, False, xformname)
-
-def read_raw_and_xform_one_object(time_spec, service_name, bucketname,\
-        prefix_in, xformname):
-    endpoint = "https://" + time_spec + ".s3proxy." + service_name + ".com:443/";
+def run_xform_periodically(seconds, service_name, bucketname, prefix_in, xformname, **kwargs):
+    endpoint = "https://s3proxy." + service_name + ".com:443/";
+    print('infinstor proxy endpoint=' + endpoint)
     session = boto3.session.Session(profile_name='infinstor')
     client = session.client('s3', endpoint_url=endpoint)
 
     xform_obj = client.get_object(Bucket='infinstor-pseudo-bucket', Key='transforms/' + xformname)
     strbody = xform_obj['Body']
     bts = strbody.read()
-    s = str(bts, 'utf-8')
-    # print(s)
+    transform_string = str(bts, 'utf-8')
+    return run_xform_string_periodically(seconds, service_name, bucketname, prefix_in, xformname, transform_string, **kwargs)
 
-    transformAst = infin_ast.extract_transform(TRANSFORM_ONE_OBJ, src_str=s)
+def run_xform_string_periodically(seconds, service_name, bucketname, prefix_in, xformname, transform_string, **kwargs):
+    while (True):
+        start_time = datetime.utcnow()
+        time.sleep(seconds)
+        end_time = datetime.utcnow()
+        time_spec = 'tm' + start_time.strftime("%Y%m%d%H%M%S")\
+                + '-tm' + end_time.strftime("%Y%m%d%H%M%S")
+        read_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, False,\
+                xformname, transform_string, False, **kwargs)
+
+def new_dict_with_just_one(dict_in):
+    one_dir = list(dict_in.keys())[0]
+    newdict = dict()
+    newdict[one_dir] = dict_in[one_dir]
+    return newdict
+
+def read_and_xform_one_object(time_spec, service_name, bucketname,\
+        prefix_in, xformname, transform_string, one_object_only, **kwargs):
+    endpoint = "https://" + time_spec + ".s3proxy." + service_name + ".com:443/";
+    session = boto3.session.Session(profile_name='infinstor')
+    client = session.client('s3', endpoint_url=endpoint)
+
+    transformAst = infin_ast.extract_transform(TRANSFORM_ONE_OBJ, src_str=transform_string)
     transformSrc = infin_ast.get_source(transformAst)
-    # transformSrc = transformSrc + "\ninfin_transform_one_object(bucket, key, object_bytes)\n"
     tree = ast.parse(transformSrc)
     # astpretty.pprint(tree)
     compiledcode1 = compile(tree, "<string>", "exec")
 
     dict_of_arrays_of_files = dict()
     list_dir_by_dir(client, bucketname, prefix_in, True, dict_of_arrays_of_files)
+    if (one_object_only): # truncate to first directory
+        dict_of_arrays_of_files = new_dict_with_just_one(dict_of_arrays_of_files)
     try:
         start_run()
         for parentdir in dict_of_arrays_of_files:
@@ -454,12 +606,93 @@ def read_raw_and_xform_one_object(time_spec, service_name, bucketname,\
             # pprint.pprint(glb)
             info("Execution of global statics for parentdir " + parentdir\
                     + " complete. status_str=" + status_str)
+            tdir = tempfile.mkdtemp()
             objects = download_objects_inner(endpoint, bucketname, parentdir, array_of_files,\
-                    False, glb)
+                        False, glb, tdir, **kwargs)
+            shutil.rmtree(tdir)
     finally:
         end_run()
     return objects
 
+def read_and_xform_dir_by_dir(time_spec, service_name, bucketname,\
+        prefix_in, xformname, transform_string, one_object_only, **kwargs):
+    endpoint = "https://" + time_spec + ".s3proxy." + service_name + ".com:443/";
+    session = boto3.session.Session(profile_name='infinstor')
+    client = session.client('s3', endpoint_url=endpoint)
+
+    transformAst = infin_ast.extract_transform(TRANSFORM_DIR_BY_DIR, src_str=transform_string)
+    transformSrc = infin_ast.get_source(transformAst)
+    tree = ast.parse(transformSrc)
+    # astpretty.pprint(tree)
+    compiledcode1 = compile(tree, "<string>", "exec")
+
+    dict_of_arrays_of_files = dict()
+    list_dir_by_dir(client, bucketname, prefix_in, True, dict_of_arrays_of_files)
+    if (one_object_only): # truncate to first directory
+        dict_of_arrays_of_files = new_dict_with_just_one(dict_of_arrays_of_files)
+    try:
+        start_run()
+        for parentdir in dict_of_arrays_of_files:
+            array_of_files = dict_of_arrays_of_files[parentdir]
+            temp_input_dir = tempfile.mkdtemp()
+            temp_output_dir = tempfile.mkdtemp()
+            info('Processing parentdir ' + parentdir + ' with ' + str(len(array_of_files))\
+                    + ' objects, temp_input_dir=' + str(temp_input_dir)\
+                    + ', temp_output_dir=' + str(temp_output_dir))
+            objects = download_objects_inner(endpoint, bucketname, parentdir, array_of_files,\
+                    False, None, temp_input_dir, **kwargs)
+
+            transformAst = infin_ast.extract_transform(TRANSFORM_DIR_BY_DIR,\
+                    src_str=transform_string)
+            transformSrc = infin_ast.get_source(transformAst)
+            xn = TRANSFORM_DIR_BY_DIR
+            tree = ast.parse(transformSrc)
+            compiledcode3 = compile(tree, "<string>", "exec")
+
+            # Add all functions in xformcode to the globals dictionary
+            glb = {}
+            fl = FuncLister(glb)
+            fl.visit(tree)
+            try:
+                exec(compiledcode3, glb)
+            except Exception as e:
+                status_str = str(e)
+                shutil.rmtree(temp_input_dir)
+                continue # process next directory
+            else:
+                status_str = 'Success'
+            # print('Globals=')
+            # pprint.pprint(glb)
+            info("Execution of global statics complete. status_str=" + status_str)
+
+            namespaced_infin_transform_fnx = namespaced_function(glb[xn], glb, None, True)
+            try:
+                namespaced_infin_transform_fnx(temp_input_dir, temp_output_dir, **kwargs)
+            except Exception as e:
+                status_str = str(e)
+                info("Error executing " + xn + ", status=" + status_str)
+            else:
+                info("Successfully executed transform " + xn + " for parentdir " + parentdir)
+            finally:
+                shutil.rmtree(temp_input_dir)
+
+            try:
+                dest_path = 'infinstor/' + parentdir
+                for one_output_file in os.listdir(temp_output_dir):
+                    fq_local = os.path.join(temp_output_dir, one_output_file)
+                    info("Logging artifact " + str(fq_local) + " to " + dest_path)
+                    log_artifact(fq_local, dest_path)
+            except Exception as e:
+                status_str = str(e)
+                info("Error logging artifacts for parentdir " + str(parentdir)\
+                        + ": " + status_str)
+            else:
+                info("Successfully logged artifacts for parentdir " + str(parentdir))
+            finally:
+                shutil.rmtree(temp_output_dir)
+    finally:
+        end_run()
+
 def look_for_transform(transform_string, transform_symbol):
     transformAst = infin_ast.extract_transform(transform_symbol, src_str=transform_string)
     transformSrc = infin_ast.get_source(transformAst)
@@ -474,7 +707,7 @@ def look_for_transform(transform_string, transform_symbol):
             return True
     return False
 
-def run_transform_inline(time_spec, service_name, bucketname, prefix_in, xformname):
+def run_transform_inline(time_spec, service_name, bucketname, prefix_in, xformname, **kwargs):
     endpoint = "https://" + time_spec + ".s3proxy." + service_name + ".com:443/";
     print('infinstor proxy endpoint=' + endpoint)
     session = boto3.session.Session(profile_name='infinstor')
@@ -484,22 +717,40 @@ def run_transform_inline(time_spec, service_name, bucketname, prefix_in, xformna
     strbody = xform_obj['Body']
     bts = strbody.read()
     transform_string = str(bts, 'utf-8')
-    # print(s)
+    return run_transform_string_inline(time_spec, service_name, bucketname, prefix_in,\
+            xformname, transform_string, False, **kwargs)
 
+def run_transform_string_inline(time_spec, service_name, bucketname, prefix_in,\
+        xformname, transform_string, one_object_only, **kwargs):
     if (look_for_transform(transform_string, TRANSFORM_RAW_PD)):
-        return read_raw_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, False, xformname)
+        print("Running transform " + str(TRANSFORM_RAW_PD))
+        return read_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, False,\
+                xformname, transform_string, one_object_only, **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_RAW_DS)):
-        return read_raw_and_xform_to_ds(time_spec, service_name, bucketname, prefix_in, False, xformname)
+        print("Running transform " + str(TRANSFORM_RAW_DS))
+        return read_and_xform_to_ds(time_spec, service_name, bucketname, prefix_in, False,\
+                xformname, transform_string, one_object_only, **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_CSV_PD)):
-        return read_raw_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, True, xformname)
+        print("Running transform " + str(TRANSFORM_CSV_PD))
+        return read_and_xform_to_pd(time_spec, service_name, bucketname, prefix_in, True,\
+                xformname, transform_string, one_object_only, **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_CSV_DS)):
-        return read_raw_and_xform_to_ds(time_spec, service_name, bucketname, prefix_in, True, xformname)
+        print("Running transform " + str(TRANSFORM_CSV_DS))
+        return read_and_xform_to_ds(time_spec, service_name, bucketname, prefix_in, True,\
+                xformname, transform_string, one_object_only, **kwargs)
     elif (look_for_transform(transform_string, TRANSFORM_ONE_OBJ)):
-        return read_raw_and_xform_one_object(time_spec, service_name, bucketname, prefix_in, xformname)
+        print("Running transform " + str(TRANSFORM_ONE_OBJ))
+        return read_and_xform_one_object(time_spec, service_name, bucketname, prefix_in,\
+                xformname, transform_string, one_object_only, **kwargs)
+    elif (look_for_transform(transform_string, TRANSFORM_DIR_BY_DIR)):
+        print("Running transform " + str(TRANSFORM_DIR_BY_DIR))
+        return read_and_xform_dir_by_dir(time_spec, service_name, bucketname, prefix_in,\
+                xformname, transform_string, one_object_only, **kwargs)
     else:
         raise Exception("Cannot find a known transform function");
 
-def run_transform_singlevm(time_spec, service_name, bucketname, prefix_in, xformname):
+def run_transform_singlevm(time_spec, service_name, bucketname, prefix_in, xformname,\
+        instance_type, **kwargs):
     # save the conda environment
     projdir = tempfile.mkdtemp()
     print('Project dir: ' + projdir)
@@ -514,6 +765,9 @@ def run_transform_singlevm(time_spec, service_name, bucketname, prefix_in, xform
     process.wait()
 
     with open(projdir + sep + 'MLproject', "w") as projfile:
+        kwp = ''
+        for key, value in kwargs.items():
+            kwp = kwp + (' --' + key + '={' + key + '}')
         projfile.write('Name: run-' + xformname + '\n')
         projfile.write('conda_env: conda.yaml\n')
         projfile.write('\n')
@@ -525,18 +779,26 @@ def run_transform_singlevm(time_spec, service_name, bucketname, prefix_in, xform
         projfile.write('      bucket: string\n')
         projfile.write('      prefix: string\n')
         projfile.write('      xformname: string\n')
+        for key, value in kwargs.items():
+            projfile.write('      ' + key + ': string\n')
         projfile.write('    command: "python -c \'from infinstor import mlflow_run; mlflow_run.main()\'\
                     --timespec={timespec} --service={service}\
-                    --bucket={bucket} --prefix={prefix} --xformname={xformname}"\n')
+                    --bucket={bucket} --prefix={prefix} --xformname={xformname}' + kwp + '"\n')
 
     child_env = os.environ.copy()
     child_env['MLFLOW_TRACKING_URI'] = 'infinstor://' + service_name + '/'
-    cmd = ['mlflow', 'run', '-b', 'infinstor-backend', projdir,\
+    cmd = ['mlflow', 'run',\
+            '-b', 'infinstor-backend', '--backend-config',\
+            '{"instance_type": "' + instance_type + '"}',\
+            projdir,\
             '-P', 'timespec=' + time_spec,\
             '-P', 'service=' + service_name,\
             '-P', 'bucket=' + bucketname,\
             '-P', 'prefix=' + prefix_in,\
             '-P', 'xformname=' + xformname ]
+    for key, value in kwargs.items():
+        cmd.append('-P')
+        cmd.append(key + '=' + value)
     print('Running cmd:')
     print(*cmd)
     process = subprocess.Popen(cmd, env=child_env, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,\
@@ -547,14 +809,39 @@ def run_transform_singlevm(time_spec, service_name, bucketname, prefix_in, xform
 
     return None
 
-def run_transform_emr(time_spec, service_name, bucketname, prefix_in, xformname):
+def run_transform_emr(time_spec, service_name, bucketname, prefix_in, xformname, **kwargs):
     print('Error: Unimplemented')
     return None
 
-def run_transform(time_spec, service_name, bucketname, prefix_in, xformname, run_location):
+def run_transform(time_spec, service_name, bucketname, prefix_in, xformname, run_location,**kwargs):
     if (run_location == "inline"):
-        return run_transform_inline(time_spec, service_name, bucketname, prefix_in, xformname)
+        return run_transform_inline(time_spec, service_name, bucketname, prefix_in, xformname,\
+                **kwargs)
     elif (run_location == "singlevm"):
-        return run_transform_singlevm(time_spec, service_name, bucketname, prefix_in, xformname)
+        instance_type = kwargs.pop('instance_type')
+        return run_transform_singlevm(time_spec, service_name, bucketname, prefix_in, xformname,\
+                instance_type, **kwargs)
     elif (run_location == "emr"):
-        return run_transform_emr(time_spec, service_name, bucketname, prefix_in, xformname)
+        return run_transform_emr(time_spec, service_name, bucketname, prefix_in, xformname,\
+                **kwargs)
+
+def test_once(servicename, bucketname, path_in_bucket,  time_spec, **kwargs):
+    caller_globals = dict(inspect.getmembers(inspect.stack()[1][0]))["f_globals"]
+    code_str = ''
+    for one_line in caller_globals['In']:
+        code_str = one_line # pick last cell's code
+    # print(code_str)
+
+    clean_code = ''
+    code_str_lines = code_str.splitlines()
+    for one_line in code_str_lines:
+        if (one_line == '%reset -f'):
+            continue
+        if (one_line == 'from infinstor import test_once # infinstor'):
+            continue
+        if (one_line.startswith('rv = test_once')):
+            continue
+        clean_code += (one_line + '\n')
+    # print(clean_code)
+    return run_transform_string_inline(time_spec, servicename, bucketname, path_in_bucket,\
+            None, clean_code, True, **kwargs)
